@@ -26,6 +26,7 @@ Vmem type_info_arena;
 static double compiler_init_time;
 static double compiler_parsing_time;
 static double compiler_sema_time;
+static double compiler_exec_time;
 static double compiler_ir_gen_time;
 static double compiler_codegen_time;
 static double compiler_link_time;
@@ -40,6 +41,8 @@ static const char *out_name(void)
 	return NULL;
 }
 
+#define START_VMEM_SIZE (sizeof(size_t) == 4 ? 1024 : 4096)
+
 void compiler_init(BuildOptions *build_options)
 {
 	// Process --path
@@ -48,6 +51,11 @@ void compiler_init(BuildOptions *build_options)
 		error_exit("Failed to change path to '%s'.", build_options->path);
 	}
 
+	FOREACH(const char *, dir, build_options->unchecked_directories)
+	{
+		(void)check_dir(dir);
+	}
+
 	compiler_init_time = -1;
 	compiler_parsing_time = -1;
 	compiler_sema_time = -1;
@@ -62,21 +70,23 @@ void compiler_init(BuildOptions *build_options)
 	//compiler.lib_dir = find_lib_dir();
 	//DEBUG_LOG("Found std library: %s", compiler.lib_dir);
 	htable_init(&compiler.context.modules, 16 * 1024);
+	pathtable_init(&compiler.context.path_symbols, INITIAL_SYMBOL_MAP);
 	decltable_init(&compiler.context.symbols, INITIAL_SYMBOL_MAP);
-	decltable_init(&compiler.context.generic_symbols, INITIAL_GENERIC_SYMBOL_MAP);
 
 	htable_init(&compiler.context.features, 1024);
 	htable_init(&compiler.context.compiler_defines, 16 * 1024);
+	methodtable_init(&compiler.context.method_extensions, 16 * 1024);
 	compiler.context.module_list = NULL;
 	compiler.context.generic_module_list = NULL;
-	compiler.context.method_extensions = NULL;
-	vmem_init(&ast_arena, 512);
+	compiler.context.method_extension_list = NULL;
+
+	vmem_init(&ast_arena, START_VMEM_SIZE);
 	ast_calloc();
-	vmem_init(&expr_arena, 512);
+	vmem_init(&expr_arena, START_VMEM_SIZE);
 	expr_calloc();
-	vmem_init(&decl_arena, 256);
+	vmem_init(&decl_arena, START_VMEM_SIZE);
 	decl_calloc();
-	vmem_init(&type_info_arena, 256);
+	vmem_init(&type_info_arena, START_VMEM_SIZE);
 	type_info_calloc();
 	// Create zero index value.
 	if (build_options->std_lib_dir)
@@ -93,6 +103,10 @@ void compiler_init(BuildOptions *build_options)
 	{
 		compiler.context.should_print_environment = true;
 	}
+	if (build_options->print_asm)
+	{
+		compiler.context.should_print_asm = true;
+	}
 }
 
 static void compiler_lex(void)
@@ -150,10 +164,12 @@ void thread_compile_task_tb(void *compile_data)
 const char *tilde_codegen(void *context)
 {
 	error_exit("TB backend not available.");
+	UNREACHABLE
 }
 void **tilde_gen(Module** modules, unsigned module_count)
 {
 	error_exit("TB backend not available.");
+	UNREACHABLE
 }
 
 #endif
@@ -161,23 +177,10 @@ void **tilde_gen(Module** modules, unsigned module_count)
 const char *build_base_name(void)
 {
 	const char *name = out_name();
-	if (!name)
-	{
-		Module **modules = compiler.context.module_list;
-		Module *main_module = (modules[0] == compiler.context.core_module && vec_size(modules) > 1) ? modules[1] : modules[0];
-		Path *path = main_module->name;
-		size_t first = 0;
-		for (size_t i = path->len; i > 0; i--)
-		{
-			if (path->module[i - 1] == ':')
-			{
-				first = i;
-				break;
-			}
-		}
-		name = &path->module[first];
-	}
-	return name;
+	if (name) return name;
+	Module **modules = compiler.context.module_list;
+	Module *main_module = (modules[0] == compiler.context.core_module && vec_size(modules) > 1) ? modules[1] : modules[0];
+	return main_module->short_path;
 }
 
 static const char *exe_name(void)
@@ -190,18 +193,15 @@ static const char *exe_name(void)
 	}
 	if (!name)
 	{
-		Path *path = compiler.context.main->unit->module->name;
-		size_t first = 0;
-		for (size_t i = path->len; i > 0; i--)
-		{
-			if (path->module[i - 1] == ':')
-			{
-				first = i;
-				break;
-			}
-		}
-		name = &path->module[first];
+		name = compiler.context.main->unit->module->short_path;
 	}
+	
+	// Use custom extension if specified
+	if (compiler.build.extension)
+	{
+		return str_cat(name, compiler.build.extension);
+	}
+	
 	switch (compiler.build.arch_os_target)
 	{
 		case WINDOWS_AARCH64:
@@ -217,6 +217,13 @@ static const char *exe_name(void)
 static const char *dynamic_lib_name(void)
 {
 	const char *name = build_base_name();
+	
+	// Use custom extension if specified
+	if (compiler.build.extension)
+	{
+		return str_cat(name, compiler.build.extension);
+	}
+	
 	switch (compiler.build.arch_os_target)
 	{
 		case WINDOWS_AARCH64:
@@ -234,6 +241,13 @@ static const char *dynamic_lib_name(void)
 static const char *static_lib_name(void)
 {
 	const char *name = build_base_name();
+	
+	// Use custom extension if specified
+	if (compiler.build.extension)
+	{
+		return str_cat(name, compiler.build.extension);
+	}
+	
 	switch (compiler.build.arch_os_target)
 	{
 		case WINDOWS_AARCH64:
@@ -306,11 +320,23 @@ static void compiler_print_bench(void)
 		double link_time = compiler_link_time - compiler_codegen_time;
 		if (compiler_link_time >= 0) last = compiler_link_time;
 		printf("Frontend -------------------- Time --- %% total\n");
-		if (compiler_init_time >= 0) printf("Initialization took: %10.3f ms  %8.1f %%\n", compiler_init_time * 1000, compiler_init_time * 100 / last);
+		if (compiler_init_time >= 0)
+		{
+			compiler_init_time -= compiler.script_time;
+			printf("Initialization took: %10.3f ms  %8.1f %%\n", compiler_init_time * 1000, compiler_init_time * 100 / last);
+			if (compiler.script_time > 0)
+			{
+				printf("Scripts took:        %10.3f ms  %8.1f %%\n", compiler.script_time * 1000, compiler.script_time * 100 / last);
+			}
+		}
 		if (compiler_parsing_time >= 0) printf("Parsing took:        %10.3f ms  %8.1f %%\n", parse_time * 1000, parse_time * 100 / last);
 		if (compiler_sema_time >= 0)
 		{
 			printf("Analysis took:       %10.3f ms  %8.1f %%\n", sema_time * 1000, sema_time * 100 / last);
+			if (compiler.exec_time > 0)
+			{
+				printf(" - Scripts took:     %10.3f ms  %8.1f %%\n", compiler.exec_time * 1000, compiler.exec_time * 100 / last);
+			}
 			printf("TOTAL:               %10.3f ms  %8.1f %%\n", compiler_sema_time * 1000, compiler_sema_time * 100 / last);
 			puts("");
 		}
@@ -320,7 +346,6 @@ static void compiler_print_bench(void)
 			printf("Ir gen took:         %10.3f ms  %8.1f %%\n", ir_time * 1000, ir_time * 100 / last);
 			if (compiler_codegen_time >= 0)
 			{
-				last = compiler_codegen_time;
 				if (compiler.build.build_threads > 1)
 				{
 					printf("Codegen took:        %10.3f ms  %8.1f %%  (%d threads)\n", codegen_time * 1000, codegen_time * 100 / last, compiler.build.build_threads);
@@ -332,7 +357,6 @@ static void compiler_print_bench(void)
 			}
 			if (compiler_link_time >= 0)
 			{
-				last = compiler_link_time;
 				printf("Linking took:        %10.3f ms  %8.1f %%\n", link_time * 1000, link_time * 100 / last);
 			}
 			printf("TOTAL:               %10.3f ms  %8.1f %%\n", (last - compiler_sema_time) * 1000, 100 - compiler_sema_time * 100 / last);
@@ -391,13 +415,32 @@ void compiler_parse(void)
 		if (!parse_stdin()) has_error = true;
 	}
 
-	if (has_error) exit_compiler(EXIT_FAILURE);
+	if (has_error)
+	{
+		if (compiler.build.lsp_output)
+		{
+			eprintf("> ENDLSP-ERROR\n");
+			exit_compiler(COMPILER_SUCCESS_EXIT);
+		}
+		exit_compiler(EXIT_FAILURE);
+	}
 	compiler_parsing_time = bench_mark();
 }
 
+bool compiler_should_output_file(const char *file)
+{
+	if (!vec_size(compiler.build.emit_only)) return true;
+	FOREACH(const char *, f, compiler.build.emit_only)
+	{
+		if (str_eq(file, f)) return true;
+	}
+	return false;
+}
+
 static void create_output_dir(const char *dir)
 {
 	if (!dir) return;
+	if (strlen(dir) == 0) return;
 	if (file_exists(dir))
 	{
 		if (!file_is_dir(dir)) error_exit("Output directory is not a directory %s.", dir);
@@ -405,7 +448,8 @@ static void create_output_dir(const char *dir)
 	}
 	scratch_buffer_clear();
 	scratch_buffer_append(dir);
-	if (!dir_make_recursive(scratch_buffer_to_string()))
+	dir_make_recursive(scratch_buffer_to_string());
+	if (!file_exists(dir))
 	{
 		error_exit("Failed to create directory '%s'.", dir);
 	}
@@ -425,6 +469,7 @@ void compiler_compile(void)
 		exit_compiler(COMPILER_SUCCESS_EXIT);
 	}
 	compiler_sema_time = bench_mark();
+	compiler_exec_time = compiler.exec_time;
 	Module **modules = compiler.context.module_list;
 	unsigned module_count = vec_size(modules);
 	if (module_count > MAX_MODULES)
@@ -435,9 +480,12 @@ void compiler_compile(void)
 	{
 		error_exit("No module to compile.");
 	}
-
 	if (compiler.build.output_headers)
 	{
+		if (compiler.build.header_file_dir)
+		{
+			create_output_dir(compiler.build.header_file_dir);
+		}
 		header_gen(modules, module_count);
 	}
 
@@ -450,20 +498,15 @@ void compiler_compile(void)
 	void **gen_contexts;
 	void (*task)(void *);
 
-
-	if (compiler.build.asm_file_dir || compiler.build.ir_file_dir || compiler.build.emit_object_files)
-	{
-		create_output_dir(compiler.build.build_dir);
-	}
-	if (compiler.build.ir_file_dir && (compiler.build.emit_llvm || compiler.build.test_output || compiler.build.lsp_output))
+	if ((compiler.build.emit_llvm || compiler.build.test_output || compiler.build.lsp_output))
 	{
 		create_output_dir(compiler.build.ir_file_dir);
 	}
-	if (compiler.build.asm_file_dir && compiler.build.emit_asm)
+	if (compiler.build.emit_asm)
 	{
 		create_output_dir(compiler.build.asm_file_dir);
 	}
-	if (compiler.build.object_file_dir && compiler.build.emit_object_files)
+	if (compiler.build.emit_object_files)
 	{
 		create_output_dir(compiler.build.object_file_dir);
 	}
@@ -479,6 +522,18 @@ void compiler_compile(void)
 	{
 		compiler.build.single_module = SINGLE_MODULE_ON;
 	}
+	if (compiler.build.emit_asm)
+	{
+		scratch_buffer_clear();
+		scratch_buffer_append(compiler.build.asm_file_dir);
+		dir_make_recursive(scratch_buffer_to_string());
+	}
+	if (compiler.build.emit_object_files)
+	{
+		scratch_buffer_clear();
+		scratch_buffer_append(compiler.build.object_file_dir);
+		dir_make_recursive(scratch_buffer_to_string());
+	}
 	switch (compiler.build.backend)
 	{
 		case BACKEND_C:
@@ -498,7 +553,7 @@ void compiler_compile(void)
 			task = &thread_compile_task_tb;
 			break;
 		default:
-			UNREACHABLE
+			UNREACHABLE_VOID
 	}
 	compiler_ir_gen_time = bench_mark();
 	const char *output_exe = NULL;
@@ -537,7 +592,7 @@ void compiler_compile(void)
 			case TARGET_TYPE_PREPARE:
 				break;
 			default:
-				UNREACHABLE
+				UNREACHABLE_VOID
 		}
 	}
 	if (compiler.build.emit_llvm)
@@ -620,7 +675,7 @@ void compiler_compile(void)
 		task_count--;
 		if (task_count)
 		{
-			taskqueue_run(compiler.build.build_threads > task_count ? task_count : compiler.build.build_threads, tasks);
+			taskqueue_run((int)(compiler.build.build_threads > task_count ? task_count : compiler.build.build_threads), tasks);
 		}
 	}
 	if (compiler.build.print_output)
@@ -656,24 +711,15 @@ void compiler_compile(void)
 		}
 		error_exit("Compilation produced no object files, maybe there was no code?");
 	}
+	if (vec_size(compiler.build.emit_only)) goto SKIP;
 	if (output_exe)
 	{
-		if (compiler.build.output_dir)
+		if (file_path_is_relative(output_exe))
 		{
-			create_output_dir(compiler.build.output_dir);
 			output_exe = file_append_path(compiler.build.output_dir, output_exe);
 		}
-		char *dir_path = file_get_dir(output_exe);
-		if (dir_path && strlen(dir_path) && !file_is_dir(dir_path))
-		{
-			error_exit("Can't create '%s', the directory '%s' could not be found.", output_exe, dir_path);
-		}
-
-		if (file_is_dir(output_exe))
-		{
-			error_exit("Cannot create exe with the name '%s' - there is already a directory with that name.", output_exe);
-		}
-
+		;
+		file_create_folders(output_exe);
 		bool system_linker_available = link_libc() && compiler.platform.os != OS_TYPE_WIN32;
 		bool use_system_linker = system_linker_available && compiler.build.arch_os_target == default_target;
 		switch (compiler.build.linker_type)
@@ -744,17 +790,26 @@ void compiler_compile(void)
 				scratch_buffer_append(name);
 			}
 			name = scratch_buffer_to_string();
+			const char *full_path = realpath(scratch_buffer_to_string(), NULL);
+			if (!full_path)
+			{
+				error_exit("The binary '%s' was unexpectedly not found.", scratch_buffer_to_string());
+			}
 			OUTF("Launching %s", name);
 			FOREACH(const char *, arg, compiler.build.args)
 			{
 				OUTF(" %s", arg);
 			}
+			if (compiler.build.run_dir)
+			{
+				OUTF(" from directory %s", compiler.build.run_dir);
+				dir_change(compiler.build.run_dir);
+			}
 			OUTN("");
-
-			int ret = run_subprocess(name, compiler.build.args);
+			int ret = run_subprocess(full_path, compiler.build.args);
 			if (compiler.build.delete_after_run)
 			{
-				file_delete_file(name);
+				file_delete_file(full_path);
 			}
 			if (ret < 0) exit_compiler(EXIT_FAILURE);
 			OUTF("Program completed with exit code %d.\n", ret);
@@ -763,20 +818,11 @@ void compiler_compile(void)
 	}
 	else if (output_static)
 	{
-		if (compiler.build.output_dir)
+		if (file_path_is_relative(output_static))
 		{
-			create_output_dir(compiler.build.output_dir);
 			output_static = file_append_path(compiler.build.output_dir, output_static);
 		}
-		char *dir_path = file_get_dir(output_static);
-		if (dir_path && strlen(dir_path) && !file_is_dir(dir_path))
-		{
-			error_exit("Can't create '%s', the directory '%s' could not be found.", output_static, dir_path);
-		}
-		if (file_is_dir(output_static))
-		{
-			error_exit("Cannot create a static library with the name '%s' - there is already a directory with that name.", output_exe);
-		}
+		file_create_folders(output_static);
 		if (!static_lib_linker(output_static, obj_files, output_file_count))
 		{
 			error_exit("Failed to produce static library '%s'.", output_static);
@@ -788,20 +834,11 @@ void compiler_compile(void)
 	}
 	else if (output_dynamic)
 	{
-		if (compiler.build.output_dir)
+		if (file_path_is_relative(output_dynamic))
 		{
-			create_output_dir(compiler.build.output_dir);
 			output_dynamic = file_append_path(compiler.build.output_dir, output_dynamic);
 		}
-		char *dir_path = file_get_dir(output_dynamic);
-		if (dir_path && strlen(dir_path) && !file_is_dir(dir_path))
-		{
-			error_exit("Can't create '%s', the directory '%s' could not be found.", output_dynamic, dir_path);
-		}
-		if (file_is_dir(output_dynamic))
-		{
-			error_exit("Cannot create a dynamic library with the name '%s' - there is already a directory with that name.", output_exe);
-		}
+		file_create_folders(output_dynamic);
 		if (!dynamic_lib_linker(output_dynamic, obj_files, output_file_count))
 		{
 			error_exit("Failed to produce dynamic library '%s'.", output_dynamic);
@@ -813,6 +850,7 @@ void compiler_compile(void)
 	}
 	else
 	{
+		SKIP:
 		compiler_print_bench();
 	}
 	free(obj_files);
@@ -821,8 +859,8 @@ INLINE void expand_csources(const char *base_dir, const char **source_dirs, cons
 {
 	if (source_dirs)
 	{
-		static const char* c_suffix_list[3] = { ".c" };
-		*sources_ref = target_expand_source_names(base_dir, source_dirs, c_suffix_list, NULL, 1, false);
+		static const char* c_suffix_list[3] = { ".c", ".m" };
+		*sources_ref = target_expand_source_names(base_dir, source_dirs, c_suffix_list, NULL, 2, false);
 	}
 }
 
@@ -863,9 +901,9 @@ void compile_file_list(BuildOptions *options)
 		{
 			error_exit("The target is a 'prepare' target, and only 'build' can be used with it.");
 		}
-		OUTF("] Running prepare target '%s'.\n", options->target_select);
+		OUTF("Running prepare target '%s'.\n", options->target_select);
 		execute_scripts();
-		OUTF("] Completed.\n.");
+		OUTN("Completed.\n.");
 		return;
 	}
 	if (options->command == COMMAND_CLEAN_RUN)
@@ -908,16 +946,7 @@ static void setup_bool_define(const char *id, bool value)
 	setup_define(id, expr_new_const_bool(INVALID_SPAN, type_bool, value));
 }
 
-#if FETCH_AVAILABLE
-const char * vendor_fetch_single(const char* lib, const char* path) 
-{
-	const char *resource = str_printf("/c3lang/vendor/releases/download/latest/%s.c3l", lib);
-	const char *destination = file_append_path(path, str_printf("%s.c3l", lib));
-	const char *error = download_file("https://github.com", resource, destination);
-	return error;	
-}
-
-static bool use_ansi(void)
+bool use_ansi(void)
 {
 	switch (compiler.context.ansi)
 	{
@@ -935,11 +964,20 @@ static bool use_ansi(void)
 #endif
 }
 
+#if FETCH_AVAILABLE
+const char * vendor_fetch_single(const char* lib, const char* path) 
+{
+	const char *resource = str_printf("/c3lang/vendor/releases/download/latest/%s.c3l", lib);
+	const char *destination = file_append_path(path, str_printf("%s.c3l", lib));
+	const char *error = download_file("https://github.com", resource, destination);
+	return error;	
+}
+
 #define PROGRESS_BAR_LENGTH 35
 
 void update_progress_bar(const char* lib, int current_step, int total_steps)
 {
-    float progress = (float)current_step / total_steps;
+    float progress = (float)current_step / (float)total_steps;
     int filled_length = (int)(progress * PROGRESS_BAR_LENGTH);
 	printf("\033[2K%-10s ", lib);
     printf("[");
@@ -948,7 +986,7 @@ void update_progress_bar(const char* lib, int current_step, int total_steps)
 	    printf(i < filled_length ? "=" : " ");
     }
 	printf("] %d%%\r", (int)(progress * 100));
-	fflush(stdout);
+	(void)fflush(stdout);
 }
 
 void vendor_fetch(BuildOptions *options)
@@ -961,14 +999,14 @@ void vendor_fetch(BuildOptions *options)
 		if (file_exists(PROJECT_JSON5) || file_exists(PROJECT_JSON))
 		{
 			const char** deps_dirs =  get_project_dependency_directories();
-			int num_lib = vec_size(deps_dirs); 
+			int num_lib = (int)vec_size(deps_dirs);
 			if (num_lib > 0) options->vendor_download_path = deps_dirs[0];
 		}
 	}
 
 	unsigned count = 0;
 	const char** fetched_libraries = NULL;
-	int total_libraries = vec_size(options->libraries_to_fetch);
+	int total_libraries = (int)vec_size(options->libraries_to_fetch);
 	
 	for(int i = 0; i < total_libraries; i++)
 	{
@@ -976,7 +1014,7 @@ void vendor_fetch(BuildOptions *options)
 		if (!ansi || total_libraries == 1)
 		{
 			printf("Fetching library '%s'...", lib);
-			fflush(stdout);
+			(void)fflush(stdout);
 		}
 		else
 		{
@@ -1007,7 +1045,7 @@ void vendor_fetch(BuildOptions *options)
 			{
 				printf("Failed: '%s'\n", error);
 			}
-			fflush(stdout);
+			(void)fflush(stdout);
 		}
 	}
 
@@ -1214,6 +1252,7 @@ void execute_scripts(void)
 			error_exit("Failed to open script dir '%s'", compiler.build.script_dir);
 		}
 	}
+	double start = bench_mark();
 	FOREACH(const char *, exec, compiler.build.exec)
 	{
 		StringSlice execs = slice_from_string(exec);
@@ -1239,103 +1278,110 @@ PRINT_SCRIPT:;
 		}
 	}
 	dir_change(old_path);
+	compiler.script_time += bench_mark() - start;
 }
 
+static void check_address_sanitizer_options(BuildTarget *target)
+{
+	if (target->feature.sanitize_memory || target->feature.sanitize_thread)
+	{
+		error_exit("Address sanitizer cannot be used together with memory or thread sanitizer.");
+	}
+	switch (target->arch_os_target)
+	{
+		case WINDOWS_X64:
+		{
+			WinCrtLinking crt_linking = target->win.crt_linking;
+			if (crt_linking == WIN_CRT_DEFAULT)
+			{
+				error_exit("Please specify `static` or `dynamic` for `wincrt` when using address sanitizer.");
+			}
+
+			if (crt_linking == WIN_CRT_STATIC_DEBUG || crt_linking == WIN_CRT_DYNAMIC_DEBUG)
+			{
+				// We currently don't have ASan runtime libraries linked against debug CRT.
+				error_exit("Address sanitizer cannot be used when using `static-debug` or `dynamic-debug` for `wincrt`. Please use `static` or `dynamic` instead.");
+			}
+
+			WARNING("Using address sanitizer on Windows requires the sanitizer option `detect_odr_violation=0`, either set by returning it from `__asan_default_options`, or via an environment variable `ASAN_OPTIONS=detect_odr_violation=0`");
+			break;
+		}
+		case LINUX_X86:
+		case LINUX_X64:
+		case MACOS_AARCH64:
+		case MACOS_X64:
+		case FREEBSD_X86:
+		case FREEBSD_X64:
+		case NETBSD_X86:
+		case NETBSD_X64:
+			break;
+		default:
+			error_exit("Address sanitizer is only supported on Linux, FreeBSD, NetBSD, Darwin and Windows.");
+	}
+}
+
+static void check_memory_sanitizer_options(BuildTarget *target)
+{
+	if (target->feature.sanitize_thread)
+	{
+		error_exit("Memory sanitizer cannot be used together thread sanitizer.");
+	}
+	switch (target->arch_os_target)
+	{
+		case LINUX_AARCH64:
+		case LINUX_X86:
+		case LINUX_X64:
+		case FREEBSD_X86:
+		case FREEBSD_X64:
+		case NETBSD_X86:
+		case NETBSD_X64:
+			break;
+		default:
+			error_exit("Memory sanitizer is only supported on Linux, FreeBSD and NetBSD.");
+	}
+	if (target->reloc_model != RELOC_BIG_PIE)
+	{
+		error_exit("Memory sanitizer requires `PIE` relocation model.");
+	}
+}
+
+static void check_thread_sanitizer_options(BuildTarget *target)
+{
+	switch (target->arch_os_target)
+	{
+		case LINUX_AARCH64:
+		case LINUX_X64:
+		case MACOS_AARCH64:
+		case MACOS_X64:
+		case FREEBSD_X64:
+		case NETBSD_X64:
+			break;
+		default:
+			error_exit("Thread sanitizer is only supported on 64-bit Linux, NetBSD, FreeBSD and Darwin.");
+	}
+}
 static void check_sanitizer_options(BuildTarget *target)
 {
-	if (target->feature.sanitize_address)
+	if (target->feature.sanitize_address) check_address_sanitizer_options(target);
+	if (target->feature.sanitize_memory) check_memory_sanitizer_options(target);
+	if (target->feature.sanitize_thread) check_thread_sanitizer_options(target);
+
+	if (target->type == TARGET_TYPE_BENCHMARK)
 	{
-		if (target->feature.sanitize_memory || target->feature.sanitize_thread)
-		{
-			error_exit("Address sanitizer cannot be used together with memory or thread sanitizer.");
-		}
-		switch (target->arch_os_target)
-		{
-			case WINDOWS_X64:
-			{
-				WinCrtLinking crt_linking = target->win.crt_linking;
-				if (crt_linking == WIN_CRT_DEFAULT)
-				{
-					error_exit("Please specify `static` or `dynamic` for `wincrt` when using address sanitizer.");
-				}
-
-				if (crt_linking == WIN_CRT_STATIC_DEBUG || crt_linking == WIN_CRT_DYNAMIC_DEBUG)
-				{
-					// We currently don't have ASan runtime libraries linked against debug CRT.
-					error_exit("Address sanitizer cannot be used when using `static-debug` or `dynamic-debug` for `wincrt`. Please use `static` or `dynamic` instead.");
-				}
-
-				WARNING("Using address sanitizer on Windows requires the sanitizer option `detect_odr_violation=0`, either set by returning it from `__asan_default_options`, or via an environment variable `ASAN_OPTIONS=detect_odr_violation=0`");
-				break;
-			}
-			case LINUX_X86:
-			case LINUX_X64:
-			case MACOS_AARCH64:
-			case MACOS_X64:
-			case FREEBSD_X86:
-			case FREEBSD_X64:
-			case NETBSD_X86:
-			case NETBSD_X64:
-				break;
-			default:
-				error_exit("Address sanitizer is only supported on Linux, Darwin and Windows.");
-		}
-		if (target->type == TARGET_TYPE_BENCHMARK)
+		if (target->feature.sanitize_address)
 		{
 			WARNING("Running benchmarks with address sanitizer enabled!");
 		}
-	}
-	if (target->feature.sanitize_memory)
-	{
-		if (target->feature.sanitize_address || target->feature.sanitize_thread)
+		else if (target->feature.sanitize_thread)
 		{
-			error_exit("Memory sanitizer cannot be used together with address or thread sanitizer.");
+			WARNING("Running benchmarks with thread sanitizer enabled!");
 		}
-		switch (target->arch_os_target)
-		{
-			case LINUX_AARCH64:
-			case LINUX_X86:
-			case LINUX_X64:
-			case FREEBSD_X86:
-			case FREEBSD_X64:
-			case NETBSD_X86:
-			case NETBSD_X64:
-				break;
-			default:
-				error_exit("Memory sanitizer is only supported on Linux.");
-		}
-		if (target->reloc_model != RELOC_BIG_PIE)
-		{
-			error_exit("Memory sanitizer requires `PIE` relocation model.");
-		}
-		if (target->type == TARGET_TYPE_BENCHMARK)
+		else if (target->feature.sanitize_memory)
 		{
 			WARNING("Running benchmarks with memory sanitizer enabled!");
 		}
 	}
-	if (target->feature.sanitize_thread)
-	{
-		if (target->feature.sanitize_address || target->feature.sanitize_memory)
-		{
-			error_exit("Thread sanitizer cannot be used together with address or memory sanitizer.");
-		}
-		switch (target->arch_os_target)
-		{
-			case LINUX_AARCH64:
-			case LINUX_X64:
-			case MACOS_AARCH64:
-			case MACOS_X64:
-			case FREEBSD_X64:
-			case NETBSD_X64:
-				break;
-			default:
-				error_exit("Thread sanitizer is only supported on 64-bit Linux and Darwin.");
-		}
-		if (target->type == TARGET_TYPE_BENCHMARK)
-		{
-			WARNING("Running benchmarks with thread sanitizer enabled!");
-		}
-	}
+
 }
 
 const char *compiler_date_to_iso(void)
@@ -1384,9 +1430,9 @@ const char *compiler_date_to_iso(void)
 		default:
 			UNREACHABLE
 	}
-	iso[5] = month / 10 + '0';
-	iso[6] = month % 10 + '0';
-	iso[8] = comp_date[4] == ' ' ? '0' : comp_date[4];
+	iso[5] = (char)(month / 10 + '0');
+	iso[6] = (char)(month % 10 + '0');
+	iso[8] = (char)(comp_date[4] == ' ' ? '0' : comp_date[4]);
 	iso[9] = comp_date[5];
 	return iso;
 }
@@ -1421,6 +1467,11 @@ void compile()
 		print_build_env();
 		exit_compiler(COMPILER_SUCCESS_EXIT);
 	}
+	if (compiler.context.should_print_asm)
+	{
+		print_asm(&compiler.platform);
+		exit_compiler(COMPILER_SUCCESS_EXIT);
+	}
 	check_sanitizer_options(&compiler.build);
 	resolve_libraries(&compiler.build);
 	compiler.context.sources = compiler.build.sources;
@@ -1429,7 +1480,6 @@ void compile()
 		expand_csources(lib->parent->dir, lib->csource_dirs, &lib->csources);
 		expand_cinclude_dirs(lib->parent->dir, lib->cinclude_dirs, &lib->cinclude_dirs);
 	}
-	TokenType type = TOKEN_CONST_IDENT;
 	FOREACH(const char *, feature_flag, compiler.build.feature_list)
 	{
 		feature_flag = symtab_preset(feature_flag, TOKEN_CONST_IDENT);
@@ -1441,6 +1491,7 @@ void compile()
 	setup_int_define("C_LONG_SIZE", compiler.platform.width_c_long, type_int);
 	setup_int_define("C_LONG_LONG_SIZE", compiler.platform.width_c_long_long, type_int);
 	setup_int_define("REGISTER_SIZE", compiler.platform.width_register, type_int);
+	setup_int_define("MAX_VECTOR_SIZE", compiler.build.max_vector_size, type_int);
 	setup_bool_define("C_CHAR_IS_SIGNED", compiler.platform.signed_c_char);
 	setup_bool_define("PLATFORM_BIG_ENDIAN", compiler.platform.big_endian);
 	setup_bool_define("PLATFORM_I128_SUPPORTED", compiler.platform.int128);
@@ -1454,6 +1505,7 @@ void compile()
 	setup_int_define("COMPILER_SIZE_OPT_LEVEL", (uint64_t)compiler.build.optsize, type_int);
 	setup_bool_define("COMPILER_SAFE_MODE", safe_mode_enabled());
 	setup_bool_define("DEBUG_SYMBOLS", compiler.build.debug_info == DEBUG_INFO_FULL);
+	setup_bool_define("PANIC_MSG", compiler.build.feature.panic_level != PANIC_OFF);
 	setup_bool_define("BACKTRACE", compiler.build.show_backtrace != SHOW_BACKTRACE_OFF);
 #if LLVM_AVAILABLE
     setup_int_define("LLVM_VERSION", llvm_version_major, type_int);
@@ -1470,7 +1522,23 @@ void compile()
 	setup_bool_define("THREAD_SANITIZER", compiler.build.feature.sanitize_thread);
 	setup_string_define("BUILD_HASH", GIT_HASH);
 	setup_string_define("BUILD_DATE", compiler_date_to_iso());
-
+	Expr *expr_names = expr_new(EXPR_CONST, INVALID_SPAN);
+	Expr *expr_emails = expr_new(EXPR_CONST, INVALID_SPAN);
+	expr_names->const_expr.const_kind = CONST_UNTYPED_LIST;
+	expr_emails->const_expr.const_kind = CONST_UNTYPED_LIST;
+	expr_names->type = type_untypedlist;
+	expr_emails->type = type_untypedlist;
+	expr_names->resolve_status = expr_emails->resolve_status = RESOLVE_DONE;
+	FOREACH(AuthorEntry, entry, compiler.build.authors)
+	{
+		Expr *const_name = expr_new_const_string(INVALID_SPAN, entry.author); // NOLINT
+		Expr *const_email = expr_new_const_string(INVALID_SPAN, entry.email ? entry.email : ""); // NOLINT
+		vec_add(expr_names->const_expr.untyped_list, const_name);
+		vec_add(expr_emails->const_expr.untyped_list, const_email);
+	}
+	setup_define("AUTHORS", expr_names);
+	setup_define("AUTHOR_EMAILS", expr_emails);
+	setup_string_define("PROJECT_VERSION", compiler.build.version);
 	type_init_cint();
 	compiler_init_time = bench_mark();
 
@@ -1488,23 +1556,14 @@ void compile()
 		compiler_parsing_time = bench_mark();
 		emit_json();
 		exit_compiler(COMPILER_SUCCESS_EXIT);
-
-		return;
 	}
 	compiler_compile();
 }
 
-
-
-
 void global_context_add_decl(Decl *decl)
 {
 	decltable_set(&compiler.context.symbols, decl);
-}
-
-void global_context_add_generic_decl(Decl *decl)
-{
-	decltable_set(&compiler.context.generic_symbols, decl);
+	pathtable_set(&compiler.context.path_symbols, decl);
 }
 
 void linking_add_link(Linking *linking, const char *link)
@@ -1559,6 +1618,27 @@ Module *compiler_find_or_create_module(Path *module_name, const char **parameter
 	// Set up the module.
 	module = CALLOCS(Module);
 	module->name = module_name;
+	module->inlined_at = (InliningSpan) { INVALID_SPAN, NULL };
+	size_t first = 0;
+	for (size_t i = module_name->len; i > 0; i--)
+	{
+		if (module_name->module[i - 1] == ':')
+		{
+			first = i;
+			break;
+		}
+	}
+	if (!first)
+	{
+		module->short_path = module_name->module;
+	}
+	else
+	{
+		const char *name = &module_name->module[first];
+		size_t len = module_name->len - first;
+		TokenType type = TOKEN_IDENT;
+		module->short_path = symtab_add(name, len, fnv1a(name, len), &type);
+	}
 	module->stage = ANALYSIS_NOT_BEGUN;
 	module->parameters = parameters;
 	module->is_generic = vec_size(parameters) > 0;
@@ -1623,8 +1703,10 @@ File *compile_and_invoke(const char *file, const char *args, const char *stdin_d
 	const char *compiler_path = file_append_path(find_executable_path(), name);
 
 	scratch_buffer_clear();
-	if (PLATFORM_WINDOWS) scratch_buffer_append_char('"');
-	scratch_buffer_append_native_safe_path(compiler_path, strlen(compiler_path));
+#if PLATFORM_WINDOWS
+	scratch_buffer_append_char('"');
+#endif
+	scratch_buffer_append_native_safe_path(compiler_path, (int)strlen(compiler_path));
 	const char *output = "__c3exec__";
 	scratch_buffer_append(" compile -g0 --single-module=yes");
 	StringSlice slice = slice_from_string(file);
@@ -1633,11 +1715,13 @@ File *compile_and_invoke(const char *file, const char *args, const char *stdin_d
 		StringSlice file_name = slice_next_token(&slice, ';');
 		if (!file_name.len) continue;
 		scratch_buffer_append(" ");
-		scratch_buffer_append_native_safe_path(file_name.ptr, file_name.len);
+		scratch_buffer_append_native_safe_path(file_name.ptr, (int)file_name.len);
 	}
 	scratch_buffer_printf(" -o %s", output);
 	char *out;
-	if (PLATFORM_WINDOWS) scratch_buffer_append_char('"');
+#if PLATFORM_WINDOWS
+	scratch_buffer_append_char('"');
+#endif
 	if (!execute_cmd_failable(scratch_buffer_to_string(), &out, NULL, limit))
 	{
 		if (strlen(out))
@@ -1678,10 +1762,16 @@ const char *default_c_compiler(void)
 	if (cc_env && strlen(cc_env) > 0)
 	{
 		INFO_LOG("Setting CC to %s from environment variable 'C3C_CC'.", cc_env);
-		cc = strdup(cc);
+		cc = strdup(cc_env);
 		return cc;
 	}
 #if PLATFORM_WINDOWS
+	WindowsSDK *sdk = windows_get_sdk();
+
+	if (sdk && sdk->cl_path)
+	{
+		return cc = sdk->cl_path;
+	}
 	return cc = "cl.exe";
 #else
 	return cc = "cc";
@@ -1699,7 +1789,8 @@ static bool is_posix(OsType os)
 		case OS_TYPE_NETBSD:
 		case OS_TYPE_LINUX:
 		case OS_TYPE_KFREEBSD:
-		case OS_TYPE_FREE_BSD:
+		case OS_TYPE_FREEBSD:
+		case OS_TYPE_OPENBSD:
 		case OS_TYPE_SOLARIS:
 			return true;
 		case OS_TYPE_WIN32:
