@@ -277,6 +277,7 @@ LLVMValueRef llvm_emit_const_initializer(GenContext *c, ConstInitializer *const_
 			AlignSize expected_align = llvm_abi_alignment(c, element_type_llvm);
 			ConstInitializer **elements = const_init->init_array.elements;
 			ASSERT(vec_size(elements) > 0 && "Array should always have gotten at least one element.");
+			if (elements > 0 && array_type->type_kind == TYPE_FLEXIBLE_ARRAY) was_modified = true;
 			ArrayIndex current_index = 0;
 			unsigned alignment = 0;
 			LLVMValueRef *parts = NULL;
@@ -371,15 +372,15 @@ LLVMValueRef llvm_emit_const_initializer(GenContext *c, ConstInitializer *const_
 			ByteSize prev_size = 0;
 			for (ArrayIndex i = 0; i < count; i++)
 			{
-				if (members[i]->padding)
+				Decl *member = members[i];
+				if (member->padding)
 				{
-					vec_add(entries, llvm_emit_const_padding(c, members[i]->padding));
+					vec_add(entries, llvm_emit_const_padding(c, member->padding));
 				}
-				LLVMTypeRef expected_type = llvm_get_type(c, const_init->init_struct[i]->type);
 				LLVMValueRef element = llvm_emit_const_initializer(c, const_init->init_struct[i]);
 				LLVMTypeRef element_type = LLVMTypeOf(element);
 				//ASSERT(LLVMIsConstant(element));
-				if (expected_type != element_type)
+				if (llvm_get_type(c, member->type) != element_type)
 				{
 					was_modified = true;
 				}
@@ -392,11 +393,11 @@ LLVMValueRef llvm_emit_const_initializer(GenContext *c, ConstInitializer *const_
 					// What is the expected offset we would get?
 					ByteSize new_offset = is_packed ? old_offset + prev_size : aligned_offset(old_offset + prev_size, llvm_abi_alignment(c, element_type));
 					// Add the padding we have built in.
-					new_offset += members[i]->padding;
+					new_offset += member->padding;
 					// If this offset is too small, add const padding.
-					if (new_offset < members[i]->offset)
+					if (new_offset < member->offset)
 					{
-						vec_add(entries, llvm_emit_const_padding(c, members[i]->offset - new_offset));
+						vec_add(entries, llvm_emit_const_padding(c, member->offset - new_offset));
 					}
 				}
 				prev_size = llvm_abi_size(c, element_type);
@@ -440,14 +441,14 @@ void llvm_emit_ptr_from_array(GenContext *c, BEValue *value)
 			BEValue member;
 			llvm_emit_slice_pointer(c, value, &member);
 			llvm_value_rvalue(c, &member);
-			llvm_value_set_address(value,
-								   member.value,
-								   type_get_ptr(value->type->array.base),
-								   type_abi_alignment(value->type->array.base));
+			llvm_value_set_address(c,
+			                       value,
+			                       member.value,
+			                       type_get_ptr(value->type->array.base), type_abi_alignment(value->type->array.base));
 			return;
 		}
 		default:
-			UNREACHABLE
+			UNREACHABLE_VOID
 	}
 }
 
@@ -824,6 +825,8 @@ static void llvm_codegen_setup()
 	intrinsic_id.masked_store = lookup_intrinsic("llvm.masked.store");
 	intrinsic_id.maximum = lookup_intrinsic("llvm.maximum");
 	intrinsic_id.maxnum = lookup_intrinsic("llvm.maxnum");
+	intrinsic_id.matrix_multiply = lookup_intrinsic("llvm.matrix.multiply");
+	intrinsic_id.matrix_transpose = lookup_intrinsic("llvm.matrix.transpose");
 	intrinsic_id.memcpy = lookup_intrinsic("llvm.memcpy");
 	intrinsic_id.memcpy_inline = lookup_intrinsic("llvm.memcpy.inline");
 	intrinsic_id.memmove = lookup_intrinsic("llvm.memmove");
@@ -853,6 +856,8 @@ static void llvm_codegen_setup()
 	intrinsic_id.sqrt = lookup_intrinsic("llvm.sqrt");
 	intrinsic_id.ssub_overflow = lookup_intrinsic("llvm.ssub.with.overflow");
 	intrinsic_id.ssub_sat = lookup_intrinsic("llvm.ssub.sat");
+	intrinsic_id.smul_fixed_sat = lookup_intrinsic("llvm.smul.fix.sat");
+	intrinsic_id.threadlocal_address = lookup_intrinsic("llvm.threadlocal.address");
 	intrinsic_id.trap = lookup_intrinsic("llvm.trap");
 	intrinsic_id.trunc = lookup_intrinsic("llvm.trunc");
 	intrinsic_id.uadd_overflow = lookup_intrinsic("llvm.uadd.with.overflow");
@@ -863,6 +868,7 @@ static void llvm_codegen_setup()
 	intrinsic_id.usub_overflow = lookup_intrinsic("llvm.usub.with.overflow");
 	intrinsic_id.ushl_sat = lookup_intrinsic("llvm.ushl.sat");
 	intrinsic_id.usub_sat = lookup_intrinsic("llvm.usub.sat");
+	intrinsic_id.umul_fixed_sat = lookup_intrinsic("llvm.umul.fix.sat");
 	intrinsic_id.vector_reduce_fmax = lookup_intrinsic("llvm.vector.reduce.fmax");
 	intrinsic_id.vector_reduce_fmin = lookup_intrinsic("llvm.vector.reduce.fmin");
 	intrinsic_id.vector_reduce_smax = lookup_intrinsic("llvm.vector.reduce.smax");
@@ -983,23 +989,24 @@ static void llvm_emit_type_decls(GenContext *context, Decl *decl)
 		case NON_TYPE_DECLS:
 		case DECL_ERASED:
 		case DECL_FNTYPE:
-			UNREACHABLE;
-		case DECL_TYPEDEF:
-			if (decl->typedef_decl.is_func)
+			UNREACHABLE_VOID;
+		case DECL_TYPE_ALIAS:
+			if (decl->type_alias_decl.is_func)
 			{
 				REMINDER("Emit func typeid");
 			}
 			break;
 		case DECL_FUNC:
 			// Never directly invoked.
-			UNREACHABLE
+			UNREACHABLE_VOID
 		case DECL_INTERFACE:
 			break;
-		case DECL_DISTINCT:
+		case DECL_TYPEDEF:
 		case DECL_STRUCT:
 		case DECL_UNION:
 		case DECL_ENUM:
 		case DECL_BITSTRUCT:
+		case DECL_CONST_ENUM:
 			llvm_get_typeid(context, decl->type);
 			break;
 	}
@@ -1064,6 +1071,7 @@ static inline void llvm_optimize(GenContext *c)
 const char *llvm_codegen(void *context)
 {
 	GenContext *c = context;
+	if (!compiler_should_output_file(c->base_name)) return NULL;
 	llvm_optimize(c);
 
 	// Serialize the LLVM IR, if requested, also verify the IR in this case
@@ -1345,7 +1353,7 @@ LLVMValueRef llvm_get_ref(GenContext *c, Decl *decl)
 			llvm_append_function_attributes(c, decl);
 			llvm_set_decl_linkage(c, decl);
 			return backend_ref;
-		case DECL_DEFINE:
+		case DECL_ALIAS:
 			return llvm_get_ref(c, decl->define_decl.alias);
 		case DECL_FAULT:
 			return llvm_create_fault(c, decl);
@@ -1353,14 +1361,16 @@ LLVMValueRef llvm_get_ref(GenContext *c, Decl *decl)
 		case DECL_ATTRIBUTE:
 		case DECL_BITSTRUCT:
 		case DECL_CT_ASSERT:
-		case DECL_DISTINCT:
+		case DECL_TYPEDEF:
 		case DECL_ENUM:
+		case DECL_CONST_ENUM:
 		case DECL_ENUM_CONSTANT:
 		case DECL_IMPORT:
+		case DECL_ALIAS_PATH:
 		case DECL_LABEL:
 		case DECL_MACRO:
 		case DECL_STRUCT:
-		case DECL_TYPEDEF:
+		case DECL_TYPE_ALIAS:
 		case DECL_UNION:
 		case DECL_DECLARRAY:
 		case DECL_BODYPARAM:
