@@ -96,6 +96,7 @@ typedef struct GenContext_
 	const char *ir_filename;
 	const char *object_filename;
 	const char *asm_filename;
+	const char *base_name;
 	LLVMTypeRef bool_type;
 	LLVMTypeRef byte_type;
 	LLVMTypeRef introspect_type;
@@ -199,6 +200,8 @@ typedef struct
 	unsigned masked_store;
 	unsigned maximum;
 	unsigned maxnum;
+	unsigned matrix_multiply;
+	unsigned matrix_transpose;
 	unsigned memcpy;
 	unsigned memcpy_inline;
 	unsigned memmove;
@@ -222,6 +225,7 @@ typedef struct
 	unsigned sin;
 	unsigned smax;
 	unsigned smin;
+	unsigned smul_fixed_sat;
 	unsigned smul_overflow;
 	unsigned sqrt;
 	unsigned sshl_sat;
@@ -229,11 +233,13 @@ typedef struct
 	unsigned ssub_sat;
 	unsigned trap;
 	unsigned debugtrap;
+	unsigned threadlocal_address;
 	unsigned trunc;
 	unsigned uadd_overflow;
 	unsigned uadd_sat;
 	unsigned umax;
 	unsigned umin;
+	unsigned umul_fixed_sat;
 	unsigned umul_overflow;
 	unsigned ushl_sat;
 	unsigned usub_overflow;
@@ -328,8 +334,8 @@ void llvm_value_rvalue(GenContext *c, BEValue *value);
 void llvm_value_deref(GenContext *c, BEValue *value);
 void llvm_value_set(BEValue *value, LLVMValueRef llvm_value, Type *type);
 void llvm_value_set_int(GenContext *c, BEValue *value, Type *type, uint64_t i);
-void llvm_value_set_address(BEValue *value, LLVMValueRef llvm_value, Type *type, AlignSize alignment);
-void llvm_value_set_address_abi_aligned(BEValue *value, LLVMValueRef llvm_value, Type *type);
+void llvm_value_set_address(GenContext *c, BEValue *value, LLVMValueRef llvm_value, Type *type, AlignSize alignment);
+void llvm_value_set_address_abi_aligned(GenContext *c, BEValue *value, LLVMValueRef llvm_value, Type *type);
 void llvm_value_set_decl_address(GenContext *c, BEValue *value, Decl *decl);
 void llvm_value_set_decl(GenContext *c, BEValue *value, Decl *decl);
 void llvm_value_fold_optional(GenContext *c, BEValue *value);
@@ -418,7 +424,7 @@ INLINE LLVMValueRef llvm_get_struct_of_type(GenContext *c, Type *type, LLVMValue
 // -- Jumps --
 void llvm_emit_cond_br(GenContext *context, BEValue *value, LLVMBasicBlockRef then_block, LLVMBasicBlockRef else_block);
 void llvm_emit_cond_br_raw(GenContext *context, LLVMValueRef b, LLVMBasicBlockRef then_block, LLVMBasicBlockRef else_block);
-void llvm_emit_br(GenContext *c, LLVMBasicBlockRef next_block);
+bool llvm_emit_br(GenContext *c, LLVMBasicBlockRef next_block);
 void llvm_emit_jump_to_optional_exit(GenContext *c, LLVMValueRef opt_value);
 void llvm_emit_return_abi(GenContext *c, BEValue *return_value, BEValue *optional);
 void llvm_emit_return_implicit(GenContext *c);
@@ -536,7 +542,7 @@ void llvm_emit_compound_stmt(GenContext *c, Ast *ast);
 LLVMValueRef llvm_emit_const_bitstruct(GenContext *c, ConstInitializer *initializer);
 void llvm_emit_function_body(GenContext *c, Decl *decl);
 void llvm_emit_dynamic_functions(GenContext *c, Decl **funcs);
-BEValue llvm_emit_assign_expr(GenContext *c, BEValue *ref, Expr *expr, LLVMValueRef optional, bool is_init);
+BEValue llvm_emit_assign_expr(GenContext *c, BEValue *ref, Expr *ref_expr, Expr *expr, LLVMValueRef optional, bool is_init);
 INLINE void llvm_emit_exprid(GenContext *c, BEValue *value, ExprId expr);
 INLINE void llvm_emit_statement_chain(GenContext *c, AstId current);
 void llvm_emit_initialize_reference_temporary_const(GenContext *c, BEValue *ref, ConstInitializer *initializer);
