@@ -1,6 +1,8 @@
 // Copyright (c) 2022-2025 Christoffer Lerno. All rights reserved.
 // Use of this source code is governed by a LGPLv3.0
 // a copy of which can be found in the LICENSE file.
+#include <math.h>
+
 #include "sema_internal.h"
 
 
@@ -173,7 +175,7 @@ static inline bool sema_expr_analyse_swizzle(SemaContext *context, Expr *expr, b
 	{
 		Expr *arg = args[i];
 		// Analyse the expressions
-		if (!sema_analyse_expr(context, arg)) return false;
+		if (!sema_analyse_expr_rvalue(context, arg)) return false;
 		// Expect vector
 		if (!type_flat_is_vector(arg->type)) RETURN_SEMA_ERROR(arg, "A vector was expected here.");
 		// Optional-ness updated
@@ -217,7 +219,7 @@ static bool sema_expr_analyse_compare_exchange(SemaContext *context, Expr *expr)
 	Expr **args = expr->call_expr.arguments;
 	Expr *pointer = args[0];
 
-	if (!sema_analyse_expr(context, pointer)) return false;
+	if (!sema_analyse_expr_rvalue(context, pointer)) return false;
 
 	bool optional = IS_OPTIONAL(pointer);
 	Type *comp_type = type_flatten(pointer->type);
@@ -289,15 +291,29 @@ static bool sema_expr_analyse_syscall(SemaContext *context, Expr *expr)
 	return true;
 }
 
+uint64_t rand_u64()
+{
+	return ((uint64_t)rand() << 48) ^ ((uint64_t)rand() << 32) ^ ((uint64_t)rand() << 16) ^ rand();
+}
+
+bool sema_expr_analyse_rnd(SemaContext *context UNUSED, Expr *expr)
+{
+	uint64_t r = rand_u64();
+	uint64_t mantissa = r >> 11;
+	double val = (double)mantissa / (double)(1ULL << 53); // Not secure random but...
+	expr_rewrite_const_float(expr, type_double, val);
+	return true;
+}
+
 bool sema_expr_analyse_str_hash(SemaContext *context, Expr *expr)
 {
 	Expr *inner = expr->call_expr.arguments[0];
-	if (!sema_analyse_expr(context, inner)) return true;
+	if (!sema_analyse_expr_rvalue(context, inner)) return true;
 	if (!expr_is_const_string(inner))
 	{
 		RETURN_SEMA_ERROR(inner, "You need a compile time constant string to take the hash of it.");
 	}
-	uint32_t hash = fnv1a(inner->const_expr.bytes.ptr, inner->const_expr.bytes.len);
+	uint32_t hash = (uint32_t)a5hash(inner->const_expr.bytes.ptr, inner->const_expr.bytes.len, 0);
 	expr_rewrite_const_int(expr, type_uint, hash);
 	return true;
 }
@@ -306,7 +322,7 @@ bool sema_expr_analyse_str_find(SemaContext *context, Expr *expr)
 {
 	Expr *inner = expr->call_expr.arguments[0];
 	Expr *inner_find = expr->call_expr.arguments[1];
-	if (!sema_analyse_expr(context, inner) || !sema_analyse_expr(context, inner_find)) return true;
+	if (!sema_analyse_expr_rvalue(context, inner) || !sema_analyse_expr_rvalue(context, inner_find)) return true;
 	if (!expr_is_const_string(inner))
 	{
 		RETURN_SEMA_ERROR(inner, "You need a compile time constant string to search.");
@@ -325,7 +341,7 @@ bool sema_expr_analyse_str_find(SemaContext *context, Expr *expr)
 bool sema_expr_analyse_str_conv(SemaContext *context, Expr *expr, BuiltinFunction func)
 {
 	Expr *inner = expr->call_expr.arguments[0];
-	if (!sema_analyse_expr(context, inner)) return true;
+	if (!sema_analyse_expr_rvalue(context, inner)) return true;
 	if (!expr_is_const_string(inner))
 	{
 		RETURN_SEMA_ERROR(inner, "You need a compile time constant string to take convert.");
@@ -345,14 +361,14 @@ bool sema_expr_analyse_str_conv(SemaContext *context, Expr *expr, BuiltinFunctio
 			for (ArraySize i = 0; i < len; i++)
 			{
 				char c = string[i];
-				new_string[i] = char_is_upper(c) ? (c | 0x20) : c;
+				new_string[i] = (char)(char_is_upper(c) ? (c | 0x20) : c);
 			}
 			break;
 		case BUILTIN_STR_UPPER:
 			for (ArraySize i = 0; i < len; i++)
 			{
 				char c = string[i];
-				new_string[i] = char_is_lower(c) ? (c & ~0x20) : c;
+				new_string[i] = (char)(char_is_lower(c) ? (c & ~0x20) : c);
 			}
 			break;
 		default:
@@ -371,6 +387,7 @@ bool sema_expr_analyse_str_conv(SemaContext *context, Expr *expr, BuiltinFunctio
 /**
  * Interpret a UTF-8 codepoint and return an integer corresponding to the raw codepoint's value.
  * @param data_ptr pointer to the pointer which scrolls along the input data string
+ * @param inc pointer to the integer which will be incremented by the number of bytes consumed
  */
 static uint32_t utf8_to_codepoint(const char *data_ptr, int *inc)
 {
@@ -402,8 +419,6 @@ static uint32_t utf8_to_codepoint(const char *data_ptr, int *inc)
 	return result;
 }
 
-
-
 bool sema_expr_analyse_str_wide(SemaContext *context, Expr *expr, BuiltinFunction func)
 {
 	Expr **args = expr->call_expr.arguments;
@@ -417,14 +432,14 @@ bool sema_expr_analyse_str_wide(SemaContext *context, Expr *expr, BuiltinFunctio
 	if (arg_count == 2)
 	{
 		Expr *zero_term = args[1];
-		if (!sema_analyse_expr(context, zero_term)) return false;
+		if (!sema_analyse_expr_rvalue(context, zero_term)) return false;
 		if (!sema_cast_const(zero_term) || !expr_is_const_bool(zero_term))
 		{
 			RETURN_SEMA_ERROR(zero_term, "Expected a boolean value here, to determine zero termination or not.");
 		}
 		zero_terminate = zero_term->const_expr.b;
 	}
-	if (!sema_analyse_expr(context, inner)) return false;
+	if (!sema_analyse_expr_rvalue(context, inner)) return false;
 	if (!sema_cast_const(inner) && !expr_is_const_string(inner))
 	{
 		RETURN_SEMA_ERROR(inner, "You need a compile time constant string to convert to a wide string.");
@@ -437,7 +452,7 @@ bool sema_expr_analyse_str_wide(SemaContext *context, Expr *expr, BuiltinFunctio
 
 	Type *type = func == BUILTIN_WIDESTRING_16 ? type_ushort : type_uint;
 	ArraySize len = (original_len + 1) * bytes_per_unit;   // inflate to unit size +null-term
-	ConstInitializer **elements = VECNEW(ConstInitializer*, len);;
+	ConstInitializer **elements = VECNEW(ConstInitializer*, len);
 
 	const char *data = string;
 	const char *end = data + original_len;
@@ -457,7 +472,7 @@ bool sema_expr_analyse_str_wide(SemaContext *context, Expr *expr, BuiltinFunctio
 		data += increment - 1;
 		if (!from_codepoint)
 		{
-			RETURN_SEMA_ERROR(inner, "Unparseable codepoint in string.");
+			RETURN_SEMA_ERROR(inner, "Unparsable codepoint in string.");
 		}
 		if (type == type_ushort && from_codepoint & 0xFFFF0000)
 		{
@@ -517,6 +532,10 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 
 	switch (func)
 	{
+		case BUILTIN_SPRINTF:
+			return sema_expr_analyse_sprintf(context, expr, args[0], &args[1], arg_count - 1);
+		case BUILTIN_RND:
+			return sema_expr_analyse_rnd(context, expr);
 		case BUILTIN_STR_HASH:
 			return sema_expr_analyse_str_hash(context, expr);
 		case BUILTIN_STR_UPPER:
@@ -557,7 +576,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 	//    exact type size, we don't do any forced promotion.
 	for (unsigned i = 0; i < arg_count; i++)
 	{
-		if (!sema_analyse_expr(context, args[i])) return false;
+		if (!sema_analyse_expr_rvalue(context, args[i])) return false;
 		optional = optional || type_is_optional(args[i]->type);
 	}
 
@@ -576,6 +595,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 		case BUILTIN_STR_FIND:
 		case BUILTIN_WIDESTRING_16:
 		case BUILTIN_WIDESTRING_32:
+		case BUILTIN_SPRINTF:
 			UNREACHABLE
 		case BUILTIN_VECCOMPGE:
 		case BUILTIN_VECCOMPEQ:
@@ -609,7 +629,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 		case BUILTIN_OVERFLOW_SUB:
 			ASSERT(arg_count == 3);
 			if (!sema_check_builtin_args(context, args,
-						     (BuiltinArg[]) {BA_INTEGER, BA_INTEGER, BA_POINTER},
+						     (BuiltinArg[]) {BA_INTLIKE, BA_INTLIKE, BA_POINTER},
 						     3)) return false;
 			if (!sema_check_builtin_args_match(context, args, 2)) return false;
 			if (type_no_optional(args[0]->type->canonical) != type_no_optional(args[2]->type->canonical->pointer))
@@ -687,9 +707,50 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_INTLIKE}, 1)) return false;
 			rtype = args[0]->type;
 			break;
+		case BUILTIN_MATRIX_TRANSPOSE:
+		{
+			ASSERT(arg_count == 3);
+			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_VEC, BA_INTEGER, BA_INTEGER}, 3)) return false;
+			if (!sema_check_builtin_args_const(context, &args[1], 2)) return false;
+			ArraySize vec_len = type_flatten(args[0]->type)->array.len;
+			Int sum = int_mul(args[1]->const_expr.ixx, args[2]->const_expr.ixx);
+			if (!int_icomp(sum, vec_len, BINARYOP_EQ))
+			{
+				RETURN_SEMA_ERROR(args[1], "Expected row * col to equal %d.", vec_len);
+			}
+			rtype = args[0]->type;
+			break;
+		}
+		case BUILTIN_MATRIX_MUL:
+			ASSERT(arg_count == 5);
+			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_VEC, BA_VEC, BA_INTEGER, BA_INTEGER, BA_INTEGER }, 5)) return false;
+			if (!sema_check_builtin_args_const(context, &args[2], 3)) return false;
+			Type *flat1 = type_flatten(args[0]->type);
+			Type *flat2 = type_flatten(args[1]->type);
+			if (flat1->array.base != flat2->array.base)
+			{
+				RETURN_SEMA_ERROR(args[1], "Expected both matrices to be of the same type.");
+			}
+			ArraySize vec_len1 = flat1->array.len;
+			ArraySize vec_len2 = flat2->array.len;
+			Int128 sum = i128_mult(args[2]->const_expr.ixx.i, args[3]->const_expr.ixx.i);
+			args[2]->type = type_int;
+			args[3]->type = type_int;
+			if (sum.high != 0 || sum.low != vec_len1)
+			{
+				RETURN_SEMA_ERROR(args[3], "Expected outer row * inner to equal %d.", vec_len1);
+			}
+			sum = i128_mult(args[3]->const_expr.ixx.i, args[4]->const_expr.ixx.i);
+			if (sum.high != 0 || sum.low != vec_len2)
+			{
+				RETURN_SEMA_ERROR(args[4], "Expected inner * outer col to equal %d.", vec_len2);
+			}
+			rtype = type_get_vector(flat1->array.base, i128_mult(args[2]->const_expr.ixx.i, args[4]->const_expr.ixx.i).low);
+			break;
 		case BUILTIN_SAT_SHL:
 		case BUILTIN_SAT_SUB:
 		case BUILTIN_SAT_ADD:
+		case BUILTIN_SAT_MUL:
 			ASSERT(arg_count == 2);
 			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_INTLIKE, BA_INTLIKE}, 2)) return false;
 			if (!sema_check_builtin_args_match(context, args, 2)) return false;
@@ -717,11 +778,8 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			{
 				RETURN_SEMA_ERROR(args[2], "This value must be a constant.");
 			}
-			else
-			{
-				Real r = args[2]->const_expr.fxx.f;
-				if (r < 0 || r > 1) RETURN_SEMA_ERROR(args[2], "The probability must be between 0 and 1.");
-			}
+			Real r = args[2]->const_expr.fxx.f;
+			if (r < 0 || r > 1) RETURN_SEMA_ERROR(args[2], "The probability must be between 0 and 1.");
 			if (!sema_check_builtin_args_match(context, args, 2)) return false;
 			rtype = args[0]->type;
 			break;
@@ -749,6 +807,11 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_FLOATLIKE, BA_FLOATLIKE, BA_FLOATLIKE},
 						     arg_count)) return false;
 			rtype = args[0]->type;
+			if (func == BUILTIN_CEIL && expr_is_const_float(args[0]))
+			{
+				expr_rewrite_const_float(expr, rtype, ceil(args[0]->const_expr.fxx.f));
+				return true;
+			}
 			break;
 		case BUILTIN_FRAMEADDRESS:
 		case BUILTIN_RETURNADDRESS:
@@ -943,6 +1006,22 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_check_builtin_args_match(context, args, 3)) return false;
 			rtype = args[0]->type;
 			break;
+		case BUILTIN_FENCE:
+			ASSERT(arg_count == 1);
+			if (!sema_check_builtin_args(context, args, (BuiltinArg[]) {BA_INTEGER}, 1)) return false;
+			if (!sema_cast_const(args[0])) RETURN_SEMA_ERROR(args[0], "Ordering must be a compile time constant.");
+			if (!is_valid_atomicity(context, args[0])) return false;
+			switch (args[0]->const_expr.ixx.i.low)
+			{
+				case ATOMIC_NONE:
+				case ATOMIC_RELAXED:
+				case ATOMIC_UNORDERED:
+					RETURN_SEMA_ERROR(args[0], "'none', 'relaxed' and 'unordered' are not valid for fence.");
+				default:
+					break;
+			}
+			rtype = type_void;
+			break;
 		case BUILTIN_ATOMIC_LOAD:
 		{
 			ASSERT(arg_count == 3);
@@ -1031,11 +1110,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_cast_const(args[2])) RETURN_SEMA_ERROR(args[2], "'is_volatile' must be a compile time constant.");
 			if (!sema_cast_const(args[3])) RETURN_SEMA_ERROR(args[3], "Ordering must be a compile time constant.");
 			if (!is_valid_atomicity(context, args[3])) return false;
-			switch (args[3]->const_expr.ixx.i.low)
-			{
-				case ATOMIC_UNORDERED:
-					RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
-			}
+			if (args[3]->const_expr.ixx.i.low == ATOMIC_UNORDERED) RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
 			if (!sema_check_alignment_expression(context, args[4])) return false;
 			rtype = args[1]->type;
 			break;
@@ -1055,11 +1130,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_cast_const(args[2])) RETURN_SEMA_ERROR(args[2], "'is_volatile' must be a compile time constant.");
 			if (!sema_cast_const(args[3])) RETURN_SEMA_ERROR(args[3], "Ordering must be a compile time constant.");
 			if (!is_valid_atomicity(context, args[3])) return false;
-			switch (args[3]->const_expr.ixx.i.low)
-			{
-				case ATOMIC_UNORDERED:
-					RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
-			}
+			if (args[3]->const_expr.ixx.i.low == ATOMIC_UNORDERED) RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
 			if (!sema_check_alignment_expression(context, args[4])) return false;
 			rtype = args[1]->type;
 			break;
@@ -1078,11 +1149,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_cast_const(args[2])) RETURN_SEMA_ERROR(args[2], "'is_volatile' must be a compile time constant.");
 			if (!sema_cast_const(args[3])) RETURN_SEMA_ERROR(args[3], "Ordering must be a compile time constant.");
 			if (!is_valid_atomicity(context, args[3])) return false;
-			switch (args[3]->const_expr.ixx.i.low)
-			{
-				case ATOMIC_UNORDERED:
-					RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
-			}
+			if (args[3]->const_expr.ixx.i.low == ATOMIC_UNORDERED) RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
 			if (!sema_check_alignment_expression(context, args[4])) return false;
 			rtype = args[1]->type;
 			break;
@@ -1102,11 +1169,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 			if (!sema_cast_const(args[2])) RETURN_SEMA_ERROR(args[2], "'is_volatile' must be a compile time constant.");
 			if (!sema_cast_const(args[3])) RETURN_SEMA_ERROR(args[3], "Ordering must be a compile time constant.");
 			if (!is_valid_atomicity(context, args[3])) return false;
-			switch (args[3]->const_expr.ixx.i.low)
-			{
-				case ATOMIC_UNORDERED:
-					RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
-			}
+			if (args[3]->const_expr.ixx.i.low == ATOMIC_UNORDERED) RETURN_SEMA_ERROR(args[3], "'unordered' is not valid ordering.");
 			if (!sema_check_alignment_expression(context, args[4])) return false;
 			rtype = args[1]->type;
 			break;
@@ -1129,6 +1192,8 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 				case ATOMIC_ACQUIRE_RELEASE:
 				case ATOMIC_ACQUIRE:
 					RETURN_SEMA_ERROR(args[3], "'acquire' and 'acquire release' are not valid for atomic stores.");
+				default:
+					break;
 			}
 			rtype = args[1]->type;
 			break;
@@ -1142,6 +1207,7 @@ bool sema_expr_analyse_builtin_call(SemaContext *context, Expr *expr)
 		case BUILTIN_TRAP:
 		case BUILTIN_BREAKPOINT:
 		case BUILTIN_UNREACHABLE:
+		case BUILTIN_RND:
 			UNREACHABLE
 	}
 	expr->type = type_add_optional(rtype, optional);
@@ -1152,10 +1218,11 @@ static inline int builtin_expected_args(BuiltinFunction func)
 {
 	switch (func)
 	{
+		case BUILTIN_SPRINTF:
 		case BUILTIN_SYSCALL:
 		case BUILTIN_WIDESTRING_16:
 		case BUILTIN_WIDESTRING_32:
-				return -1;
+			return -1;
 		case BUILTIN_SWIZZLE:
 			return -2;
 		case BUILTIN_SWIZZLE2:
@@ -1165,6 +1232,7 @@ static inline int builtin_expected_args(BuiltinFunction func)
 		case BUILTIN_TRAP:
 		case BUILTIN_BREAKPOINT:
 		case BUILTIN_UNREACHABLE:
+		case BUILTIN_RND:
 			return 0;
 		case BUILTIN_ABS:
 		case BUILTIN_BITREVERSE:
@@ -1176,6 +1244,7 @@ static inline int builtin_expected_args(BuiltinFunction func)
 		case BUILTIN_EXACT_NEG:
 		case BUILTIN_EXP2:
 		case BUILTIN_EXP:
+		case BUILTIN_FENCE:
 		case BUILTIN_FLOOR:
 		case BUILTIN_FRAMEADDRESS:
 		case BUILTIN_LLRINT:
@@ -1226,6 +1295,7 @@ static inline int builtin_expected_args(BuiltinFunction func)
 		case BUILTIN_SAT_ADD:
 		case BUILTIN_SAT_SHL:
 		case BUILTIN_SAT_SUB:
+		case BUILTIN_SAT_MUL:
 		case BUILTIN_VOLATILE_STORE:
 		case BUILTIN_VECCOMPNE:
 		case BUILTIN_VECCOMPLT:
@@ -1249,6 +1319,7 @@ static inline int builtin_expected_args(BuiltinFunction func)
 		case BUILTIN_ATOMIC_LOAD:
 		case BUILTIN_UNALIGNED_STORE:
 		case BUILTIN_SELECT:
+		case BUILTIN_MATRIX_TRANSPOSE:
 			return 3;
 		case BUILTIN_ATOMIC_STORE:
 		case BUILTIN_MASKED_STORE:
@@ -1267,6 +1338,7 @@ static inline int builtin_expected_args(BuiltinFunction func)
 		case BUILTIN_ATOMIC_FETCH_MIN:
 		case BUILTIN_ATOMIC_FETCH_SUB:
 		case BUILTIN_ATOMIC_FETCH_DEC_WRAP:
+		case BUILTIN_MATRIX_MUL:
 			return 5;
 		case BUILTIN_MEMCOPY:
 		case BUILTIN_MEMCOPY_INLINE:
