@@ -11,7 +11,7 @@ static inline bool sema_resolve_type(SemaContext *context, TypeInfo *type_info,
 static bool sema_resolve_type_identifier(SemaContext *context, TypeInfo *type_info, ResolveTypeKind resolve_type_kind);
 INLINE bool sema_resolve_vatype(SemaContext *context, TypeInfo *type_info);
 INLINE bool sema_resolve_evaltype(SemaContext *context, TypeInfo *type_info, ResolveTypeKind resolve_kind);
-INLINE bool sema_resolve_typefrom(SemaContext *context, TypeInfo *type_info);
+INLINE bool sema_resolve_typefrom(SemaContext *context, TypeInfo *type_info, ResolveTypeKind resolve_kind);
 INLINE bool sema_resolve_typeof(SemaContext *context, TypeInfo *type_info);
 static int compare_function(Signature *sig, FunctionPrototype *proto);
 
@@ -39,7 +39,7 @@ bool sema_resolve_array_like_len(SemaContext *context, TypeInfo *type_info, Arra
 	Expr *len_expr = type_info->array.len;
 
 	// Analyse it.
-	if (!sema_analyse_expr(context, len_expr)) return type_info_poison(type_info);
+	if (!sema_analyse_expr_rvalue(context, len_expr)) return type_info_poison(type_info);
 
 	if (!cast_to_index_len(context, len_expr, true)) return type_info_poison(type_info);
 
@@ -81,17 +81,16 @@ bool sema_resolve_array_like_len(SemaContext *context, TypeInfo *type_info, Arra
 	}
 
 	// Check max values.
-	if (int_icomp(len, is_vector ? MAX_VECTOR_WIDTH : MAX_ARRAY_SIZE, BINARYOP_GT))
+	if (int_icomp(len, is_vector ? compiler.build.max_vector_size / 8 : MAX_ARRAY_SIZE, BINARYOP_GT))
 	{
 		if (is_vector)
 		{
-			SEMA_ERROR(len_expr, "A vector may not exceed %d in width.", MAX_VECTOR_WIDTH);
+			RETURN_VAL_SEMA_ERROR(type_info_poison(type_info), len_expr, "A vector may not exceed %d in bit width.", compiler.build.max_vector_size);
 		}
 		else
 		{
-			SEMA_ERROR(len_expr, "The array length may not exceed %lld.", MAX_ARRAY_SIZE);
+			RETURN_VAL_SEMA_ERROR(type_info_poison(type_info), len_expr, "The array length may not exceed %lld.", MAX_ARRAY_SIZE);
 		}
-		return type_info_poison(type_info);
 	}
 	// We're done, return the size and mark it as a success.
 	*len_ref = (ArraySize)len.i.low;
@@ -190,6 +189,15 @@ static inline bool sema_resolve_array_type(SemaContext *context, TypeInfo *type,
 	if (!sema_check_array_type(context, type, type->array.base->type, kind, len, &type->type)) return type_info_poison(type);
 	ASSERT(!type->array.len || sema_cast_const(type->array.len));
 	type->resolve_status = RESOLVE_DONE;
+	if (kind == TYPE_INFO_VECTOR)
+	{
+		if (kind == TYPE_INFO_VECTOR && type_size(type->type) > compiler.build.max_vector_size / 8)
+		{
+			RETURN_SEMA_ERROR(type, "Vectors with bitsize over %u are not supported (this vector is %llu bits), "
+						   "but you can increase the maximum allowed using '--max-vector-size'.",
+						   compiler.build.max_vector_size, (unsigned long long)type_size(type->type) * 8);
+		}
+	}
 	return true;
 
 }
@@ -226,7 +234,8 @@ static bool sema_resolve_type_identifier(SemaContext *context, TypeInfo *type_in
 			type_info->type = decl->type;
 			type_info->resolve_status = RESOLVE_DONE;
 			return true;
-		case DECL_DISTINCT:
+		case DECL_CONST_ENUM:
+		case DECL_TYPEDEF:
 			if (resolve_type_kind & RESOLVE_TYPE_NO_CHECK_DISTINCT)
 			{
 				type_info->type = decl->type;
@@ -234,9 +243,10 @@ static bool sema_resolve_type_identifier(SemaContext *context, TypeInfo *type_in
 				return true;
 			}
 			FALLTHROUGH;
-		case DECL_TYPEDEF:
+		case DECL_TYPE_ALIAS:
 			if (!sema_analyse_decl(context, decl)) return type_info_poison(type_info);
 			type_info->type = decl->type;
+			assert (type_info->type->canonical->type_kind != TYPE_ALIAS);
 			type_info->resolve_status = RESOLVE_DONE;
 			return true;
 		case DECL_POISONED:
@@ -245,33 +255,34 @@ static bool sema_resolve_type_identifier(SemaContext *context, TypeInfo *type_in
 			if (decl->var.kind == VARDECL_PARAM_CT_TYPE || decl->var.kind == VARDECL_LOCAL_CT_TYPE)
 			{
 				decl->var.is_read = true;
-				if (!decl->var.init_expr)
+				Expr *init_expr = decl->var.init_expr;
+				if (!init_expr)
 				{
-					SEMA_ERROR(type_info, "You need to assign a type to '%s' before using it.", decl->name);
-					return false;
+					RETURN_SEMA_ERROR(type_info, "You need to assign a type to '%s' before using it.", decl->name);
 				}
-				ASSERT(decl->var.init_expr->expr_kind == EXPR_TYPEINFO);
-				ASSERT(decl->var.init_expr->resolve_status == RESOLVE_DONE);
-				*type_info = *decl->var.init_expr->type_expr;
+				ASSERT_SPAN(init_expr, expr_is_const_typeid(init_expr));
+				ASSERT_SPAN(init_expr, init_expr->resolve_status == RESOLVE_DONE);
+				type_info->type = init_expr->const_expr.typeid;
 				return true;
 			}
 			FALLTHROUGH;
-		case DECL_DEFINE:
-		case DECL_FUNC:
-		case DECL_ENUM_CONSTANT:
-		case DECL_IMPORT:
-		case DECL_MACRO:
-		case DECL_LABEL:
+		case DECL_ALIAS:
+		case DECL_ALIAS_PATH:
 		case DECL_ATTRIBUTE:
+		case DECL_ENUM_CONSTANT:
 		case DECL_FAULT:
+		case DECL_FUNC:
+		case DECL_IMPORT:
+		case DECL_LABEL:
+		case DECL_MACRO:
 			SEMA_ERROR(type_info, "This is not a type.");
 			return type_info_poison(type_info);
+		case DECL_BODYPARAM:
 		case DECL_CT_ASSERT:
 		case DECL_CT_ECHO:
-		case DECL_DECLARRAY:
-		case DECL_BODYPARAM:
-		case DECL_CT_INCLUDE:
 		case DECL_CT_EXEC:
+		case DECL_CT_INCLUDE:
+		case DECL_DECLARRAY:
 		case DECL_GROUP:
 			UNREACHABLE
 	}
@@ -283,9 +294,10 @@ static bool sema_resolve_type_identifier(SemaContext *context, TypeInfo *type_in
 // $evaltype("Foo")
 INLINE bool sema_resolve_evaltype(SemaContext *context, TypeInfo *type_info, ResolveTypeKind resolve_kind)
 {
+	SEMA_DEPRECATED(type_info, "$evaltype is deprecated, use $typefrom instead.");
 	Expr *expr = type_info->unresolved_type_expr;
 	Expr *inner = sema_ct_eval_expr(context, true, expr, true);
-	if (!inner) return type_info_poison(type_info);
+	if (!inner || !expr_ok(inner)) return type_info_poison(type_info);
 	if (inner->expr_kind != EXPR_TYPEINFO)
 	{
 		SEMA_ERROR(expr, "Only type names may be resolved with $evaltype.");
@@ -316,7 +328,7 @@ INLINE bool sema_resolve_typeof(SemaContext *context, TypeInfo *type_info)
 	Expr *expr = type_info->unresolved_type_expr;
 	bool in_no_eval = context->call_env.in_no_eval;
 	context->call_env.in_no_eval = true;
-	bool success = sema_analyse_expr_value(context, expr);
+	bool success = sema_analyse_expr(context, expr);
 	context->call_env.in_no_eval = in_no_eval;
 	if (!success) return false;
 	Type *expr_type = expr->type;
@@ -328,6 +340,7 @@ INLINE bool sema_resolve_typeof(SemaContext *context, TypeInfo *type_info)
 		case STORAGE_NORMAL:
 		case STORAGE_VOID:
 		case STORAGE_UNKNOWN:
+		case STORAGE_COMPILE_TIME:
 			type_info->type = expr_type;
 			return true;
 		case STORAGE_WILDCARD:
@@ -337,22 +350,54 @@ INLINE bool sema_resolve_typeof(SemaContext *context, TypeInfo *type_info)
 				return true;
 			}
 			RETURN_SEMA_ERROR(expr, "This %sexpression lacks a concrete type.", type_is_optional(expr_type) ? "optional " : "");
-		case STORAGE_COMPILE_TIME:
-			RETURN_SEMA_ERROR(expr, "This expression has a compile time type %s.", type_quoted_error_string(expr_type));
 	}
 	UNREACHABLE
 }
 
-INLINE bool sema_resolve_typefrom(SemaContext *context, TypeInfo *type_info)
+INLINE bool sema_resolve_typefrom(SemaContext *context, TypeInfo *type_info, ResolveTypeKind resolve_kind)
 {
 	Expr *expr = type_info->unresolved_type_expr;
-	if (!sema_analyse_expr(context, expr)) return false;
-	if (!sema_cast_const(expr) || expr->const_expr.const_kind != CONST_TYPEID)
+	if (!sema_analyse_expr_rvalue(context, expr)) return false;
+	if (!sema_cast_const(expr))
 	{
-		RETURN_SEMA_ERROR(expr, "Expected a constant typeid value.");
+		RETURN_SEMA_ERROR(expr, "Expected a constant value.");
 	}
-	type_info->type = expr->const_expr.typeid;
-	return true;
+	switch (expr->const_expr.const_kind)
+	{
+		case CONST_TYPEID:
+			type_info->type = expr->const_expr.typeid;
+			return true;
+		case CONST_STRING:
+			break;
+		default:
+			RETURN_SEMA_ERROR(expr, "Expected a constant string or typeid value.");
+	}
+
+	const char *bytes = expr->const_expr.bytes.ptr;
+	ArraySize len = expr->const_expr.bytes.len;
+	Expr *typefrom = sema_resolve_string_ident(context, expr, true);
+	if (!typefrom || !expr_ok(typefrom)) return false;
+	if (typefrom->expr_kind != EXPR_TYPEINFO)
+	{
+		RETURN_SEMA_ERROR(expr, "Expected a type, not a regular identifier '%.*s'.", (int)len, bytes);
+	}
+	TypeInfo *info = typefrom->type_expr;
+	if (!sema_resolve_type(context, info, resolve_kind)) return false;
+	switch (sema_resolve_storage_type(context, info->type))
+	{
+		case STORAGE_ERROR:
+			return false;
+		case STORAGE_VOID:
+		case STORAGE_UNKNOWN:
+		case STORAGE_NORMAL:
+			type_info->type = info->type;
+			return true;
+		case STORAGE_WILDCARD:
+			RETURN_SEMA_ERROR(expr, "$typefrom failed to resolve \"%.*s\" to a definite type.", (int)len, bytes);
+		case STORAGE_COMPILE_TIME:
+			RETURN_SEMA_ERROR(expr, "$typefrom does not support compile-time types.");
+	}
+	UNREACHABLE
 }
 
 // $vatype(...)
@@ -363,7 +408,7 @@ INLINE bool sema_resolve_vatype(SemaContext *context, TypeInfo *type_info)
 		RETURN_SEMA_ERROR(type_info, "'%s' can only be used inside of a macro.", token_type_to_string(TOKEN_CT_VATYPE));
 	}
 	ASSIGN_EXPR_OR_RET(Expr *arg_expr, sema_expr_analyse_ct_arg_index(context, type_info->unresolved_type_expr, NULL), false);
-	if (!sema_analyse_expr_value(context, arg_expr)) return false;
+	if (!sema_analyse_expr(context, arg_expr)) return false;
 	if (arg_expr->expr_kind != EXPR_TYPEINFO) RETURN_SEMA_ERROR(arg_expr, "The argument was not a type.");
 	type_info->type = arg_expr->type_expr->type;
 	return true;
@@ -378,10 +423,10 @@ bool sema_unresolved_type_is_generic(SemaContext *context, TypeInfo *type_info)
 	if (type_info->subtype != TYPE_COMPRESSED_NONE) return false;
 	Decl *decl = sema_find_path_symbol(context, type_info->unresolved.name, type_info->unresolved.path);
 	if (!decl) return false;
-	if (decl->decl_kind != DECL_TYPEDEF) return false;
+	if (decl->decl_kind != DECL_TYPE_ALIAS) return false;
 	if (decl->resolve_status == RESOLVE_DONE) return false;
-	if (decl->typedef_decl.is_func) return false;
-	type_info = decl->typedef_decl.type_info;
+	if (decl->type_alias_decl.is_func) return false;
+	type_info = decl->type_alias_decl.type_info;
 	goto RETRY;
 }
 
@@ -403,15 +448,22 @@ INLINE bool sema_resolve_generic_type(SemaContext *context, TypeInfo *type_info)
 	ASSERT_SPAN(inner, inner->resolve_status == RESOLVE_NOT_DONE);
 
 	bool was_recursive = false;
+	if (compiler.generic_depth >= MAX_GENERIC_DEPTH)
+	{
+		RETURN_SEMA_ERROR(type_info, "Generic resolution of this type has become deeply nested, it was aborted after reaching %d recursions.", compiler.generic_depth);
+	}
+	compiler.generic_depth++;
 	Decl *type = sema_analyse_parameterized_identifier(context, inner->unresolved.path, inner->unresolved.name,
-	                                                   inner->span, type_info->generic.params, &was_recursive);
+	                                                   inner->span, type_info->generic.params, &was_recursive, type_info->span);
+	compiler.generic_depth--;
 	if (!decl_ok(type)) return false;
+	if (!sema_analyse_decl(context, type)) return false;
 	type_info->type = type->type;
 	if (!was_recursive) return true;
 	if (!context->current_macro && (context->call_env.kind == CALL_ENV_FUNCTION || context->call_env.kind == CALL_ENV_FUNCTION_STATIC)
 	    && !context->call_env.current_function->func_decl.in_macro)
 	{
-		RETURN_SEMA_ERROR(type_info, "Recursively generic type declarations are only allowed inside of macros. Use `def` to define an alias for the type instead.");
+		RETURN_SEMA_ERROR(type_info, "Recursively generic type declarations are only allowed inside of macros. Use `alias` to define an alias for the type instead.");
 	}
 	return true;
 }
@@ -443,6 +495,19 @@ static inline bool sema_resolve_type(SemaContext *context, TypeInfo *type_info,
 
 	type_info->resolve_status = RESOLVE_RUNNING;
 	TypeInfoCompressedKind kind = type_info->subtype;
+	switch (kind)
+	{
+		case TYPE_COMPRESSED_NONE:
+		case TYPE_COMPRESSED_PTR:
+		case TYPE_COMPRESSED_SUBPTR:
+		case TYPE_COMPRESSED_PTRPTR:
+		case TYPE_COMPRESSED_PTRSUB:
+			break;
+		case TYPE_COMPRESSED_SUB:
+		case TYPE_COMPRESSED_SUBSUB:
+			resolve_kind = resolve_kind & ~RESOLVE_TYPE_NO_CHECK_DISTINCT;
+			break;
+	}
 	switch (type_info->kind)
 	{
 		case TYPE_INFO_POISON:
@@ -465,7 +530,7 @@ static inline bool sema_resolve_type(SemaContext *context, TypeInfo *type_info,
 			if (!sema_resolve_typeof(context, type_info)) return type_info_poison(type_info);
 			goto APPEND_QUALIFIERS;
 		case TYPE_INFO_TYPEFROM:
-			if (!sema_resolve_typefrom(context, type_info)) return type_info_poison(type_info);
+			if (!sema_resolve_typefrom(context, type_info, resolve_kind)) return type_info_poison(type_info);
 			goto APPEND_QUALIFIERS;
 		case TYPE_INFO_INFERRED_VECTOR:
 		case TYPE_INFO_INFERRED_ARRAY:
@@ -554,7 +619,7 @@ static Type *flatten_raw_function_type(Type *type)
 	Type *current;
 	switch (type->type_kind)
 	{
-		case TYPE_TYPEDEF:
+		case TYPE_ALIAS:
 			return flatten_raw_function_type(type->canonical);
 		case TYPE_FUNC_RAW:
 			return type->function.prototype->raw_type;
